#ifndef _SISCONE_CONFIG_H
#define _SISCONE_CONFIG_H 1
 
/* siscone/config.h. Generated automatically at end of configure. */
/* siscone/config_raw.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define to 1 if you have the <dlfcn.h> header file. */
#ifndef SISCONE_HAVE_DLFCN_H 
#define SISCONE_HAVE_DLFCN_H  1 
#endif

/* Define to 1 if you have the <inttypes.h> header file. */
#ifndef SISCONE_HAVE_INTTYPES_H 
#define SISCONE_HAVE_INTTYPES_H  1 
#endif

/* Define to 1 if you have the `m' library (-lm). */
#ifndef SISCONE_HAVE_LIBM 
#define SISCONE_HAVE_LIBM  1 
#endif

/* Define to 1 if you have the <stdint.h> header file. */
#ifndef SISCONE_HAVE_STDINT_H 
#define SISCONE_HAVE_STDINT_H  1 
#endif

/* Define to 1 if you have the <stdio.h> header file. */
#ifndef SISCONE_HAVE_STDIO_H 
#define SISCONE_HAVE_STDIO_H  1 
#endif

/* Define to 1 if you have the <stdlib.h> header file. */
#ifndef SISCONE_HAVE_STDLIB_H 
#define SISCONE_HAVE_STDLIB_H  1 
#endif

/* Define to 1 if you have the <strings.h> header file. */
#ifndef SISCONE_HAVE_STRINGS_H 
#define SISCONE_HAVE_STRINGS_H  1 
#endif

/* Define to 1 if you have the <string.h> header file. */
#ifndef SISCONE_HAVE_STRING_H 
#define SISCONE_HAVE_STRING_H  1 
#endif

/* Define to 1 if you have the <sys/stat.h> header file. */
#ifndef SISCONE_HAVE_SYS_STAT_H 
#define SISCONE_HAVE_SYS_STAT_H  1 
#endif

/* Define to 1 if you have the <sys/types.h> header file. */
#ifndef SISCONE_HAVE_SYS_TYPES_H 
#define SISCONE_HAVE_SYS_TYPES_H  1 
#endif

/* Define to 1 if you have the <unistd.h> header file. */
#ifndef SISCONE_HAVE_UNISTD_H 
#define SISCONE_HAVE_UNISTD_H  1 
#endif

/* Define to the sub-directory where libtool stores uninstalled libraries. */
#ifndef SISCONE_LT_OBJDIR 
#define SISCONE_LT_OBJDIR  ".libs/" 
#endif

/* Name of package */
#ifndef SISCONE_PACKAGE 
#define SISCONE_PACKAGE  "siscone" 
#endif

/* Define to the address where bug reports for this package should be sent. */
#ifndef SISCONE_PACKAGE_BUGREPORT 
#define SISCONE_PACKAGE_BUGREPORT  "" 
#endif

/* Define to the full name of this package. */
#ifndef SISCONE_PACKAGE_NAME 
#define SISCONE_PACKAGE_NAME  "SISCone" 
#endif

/* Define to the full name and version of this package. */
#ifndef SISCONE_PACKAGE_STRING 
#define SISCONE_PACKAGE_STRING  "SISCone 3.0.6" 
#endif

/* Define to the one symbol short name of this package. */
#ifndef SISCONE_PACKAGE_TARNAME 
#define SISCONE_PACKAGE_TARNAME  "siscone" 
#endif

/* Define to the home page for this package. */
#ifndef SISCONE_PACKAGE_URL 
#define SISCONE_PACKAGE_URL  "" 
#endif

/* Define to the version of this package. */
#ifndef SISCONE_PACKAGE_VERSION 
#define SISCONE_PACKAGE_VERSION  "3.0.6" 
#endif

/* Define to 1 if all of the C90 standard headers exist (not just the ones
   required in a freestanding environment). This macro is provided for
   backward compatibility; new code need not use it. */
#ifndef SISCONE_STDC_HEADERS 
#define SISCONE_STDC_HEADERS  1 
#endif

/* use unique_ptr instead of auto_ptr */
#ifndef SISCONE_USES_UNIQUE_PTR_AS_AUTO_PTR 
#define SISCONE_USES_UNIQUE_PTR_AS_AUTO_PTR  /**/ 
#endif

/* Version number of package */
#ifndef SISCONE_VERSION 
#define SISCONE_VERSION  "3.0.6" 
#endif
 
/* once: _SISCONE_CONFIG_H */
#endif
